package org.apache.bcel.generic;

import java.util.Objects;
import org.apache.bcel.classfile.AnnotationEntry;
import org.apache.bcel.classfile.RuntimeInvisibleAnnotations;
import org.apache.bcel.classfile.RuntimeVisibleAnnotations;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.classfile.ConstantPool;
import java.util.Arrays;
import org.apache.commons.lang3.ArrayUtils;
import java.util.Iterator;
import org.apache.bcel.classfile.SourceFile;
import org.apache.bcel.classfile.Annotations;
import java.util.Collection;
import java.util.Collections;
import java.util.ArrayList;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Field;
import java.util.List;
import org.apache.bcel.util.BCELComparator;
import org.apache.bcel.classfile.AccessFlags;

public class ClassGen extends AccessFlags implements Cloneable
{
    private static BCELComparator bcelComparator;
    private String className;
    private String superClassName;
    private final String fileName;
    private int classNameIndex;
    private int superclassNameIndex;
    private int major;
    private int minor;
    private ConstantPoolGen cp;
    private final List<Field> fieldList;
    private final List<Method> methodList;
    private final List<Attribute> attributeList;
    private final List<String> interfaceList;
    private final List<AnnotationEntryGen> annotationList;
    private List<ClassObserver> observers;
    
    public static BCELComparator getComparator() {
        return ClassGen.bcelComparator;
    }
    
    public static void setComparator(final BCELComparator comparator) {
        ClassGen.bcelComparator = comparator;
    }
    
    public ClassGen(final JavaClass clazz) {
        super(clazz.getAccessFlags());
        this.classNameIndex = -1;
        this.superclassNameIndex = -1;
        this.major = 45;
        this.minor = 3;
        this.fieldList = new ArrayList<Field>();
        this.methodList = new ArrayList<Method>();
        this.attributeList = new ArrayList<Attribute>();
        this.interfaceList = new ArrayList<String>();
        this.annotationList = new ArrayList<AnnotationEntryGen>();
        this.classNameIndex = clazz.getClassNameIndex();
        this.superclassNameIndex = clazz.getSuperclassNameIndex();
        this.className = clazz.getClassName();
        this.superClassName = clazz.getSuperclassName();
        this.fileName = clazz.getSourceFileName();
        this.cp = new ConstantPoolGen(clazz.getConstantPool());
        this.major = clazz.getMajor();
        this.minor = clazz.getMinor();
        final Attribute[] attributes = clazz.getAttributes();
        final AnnotationEntryGen[] annotations = this.unpackAnnotations(attributes);
        Collections.addAll(this.interfaceList, clazz.getInterfaceNames());
        for (final Attribute attribute : attributes) {
            if (!(attribute instanceof Annotations)) {
                this.addAttribute(attribute);
            }
        }
        Collections.addAll(this.annotationList, annotations);
        Collections.addAll(this.methodList, clazz.getMethods());
        Collections.addAll(this.fieldList, clazz.getFields());
    }
    
    public ClassGen(final String className, final String superClassName, final String fileName, final int accessFlags, final String[] interfaces) {
        this(className, superClassName, fileName, accessFlags, interfaces, new ConstantPoolGen());
    }
    
    public ClassGen(final String className, final String superClassName, final String fileName, final int accessFlags, final String[] interfaces, final ConstantPoolGen cp) {
        super(accessFlags);
        this.classNameIndex = -1;
        this.superclassNameIndex = -1;
        this.major = 45;
        this.minor = 3;
        this.fieldList = new ArrayList<Field>();
        this.methodList = new ArrayList<Method>();
        this.attributeList = new ArrayList<Attribute>();
        this.interfaceList = new ArrayList<String>();
        this.annotationList = new ArrayList<AnnotationEntryGen>();
        this.className = className;
        this.superClassName = superClassName;
        this.fileName = fileName;
        this.cp = cp;
        if (fileName != null) {
            this.addAttribute(new SourceFile(cp.addUtf8("SourceFile"), 2, cp.addUtf8(fileName), cp.getConstantPool()));
        }
        this.classNameIndex = cp.addClass(className);
        this.superclassNameIndex = cp.addClass(superClassName);
        if (interfaces != null) {
            Collections.addAll(this.interfaceList, interfaces);
        }
    }
    
    public void addAnnotationEntry(final AnnotationEntryGen a) {
        this.annotationList.add(a);
    }
    
    public void addAttribute(final Attribute a) {
        this.attributeList.add(a);
    }
    
    public void addEmptyConstructor(final int accessFlags) {
        final InstructionList il = new InstructionList();
        il.append(InstructionConst.THIS);
        il.append(new INVOKESPECIAL(this.cp.addMethodref(this.superClassName, "<init>", "()V")));
        il.append(InstructionConst.RETURN);
        final MethodGen mg = new MethodGen(accessFlags, Type.VOID, Type.NO_ARGS, null, "<init>", this.className, il, this.cp);
        mg.setMaxStack(1);
        this.addMethod(mg.getMethod());
    }
    
    public void addField(final Field f) {
        this.fieldList.add(f);
    }
    
    public void addInterface(final String name) {
        this.interfaceList.add(name);
    }
    
    public void addMethod(final Method m) {
        this.methodList.add(m);
    }
    
    public void addObserver(final ClassObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList<ClassObserver>();
        }
        this.observers.add(o);
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (final CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
    }
    
    public boolean containsField(final Field f) {
        return this.fieldList.contains(f);
    }
    
    public Field containsField(final String name) {
        for (final Field f : this.fieldList) {
            if (f.getName().equals(name)) {
                return f;
            }
        }
        return null;
    }
    
    public Method containsMethod(final String name, final String signature) {
        for (final Method m : this.methodList) {
            if (m.getName().equals(name) && m.getSignature().equals(signature)) {
                return m;
            }
        }
        return null;
    }
    
    @Override
    public boolean equals(final Object obj) {
        return ClassGen.bcelComparator.equals(this, obj);
    }
    
    public AnnotationEntryGen[] getAnnotationEntries() {
        return this.annotationList.toArray(AnnotationEntryGen.EMPTY_ARRAY);
    }
    
    public Attribute[] getAttributes() {
        return this.attributeList.toArray(Attribute.EMPTY_ARRAY);
    }
    
    public String getClassName() {
        return this.className;
    }
    
    public int getClassNameIndex() {
        return this.classNameIndex;
    }
    
    public ConstantPoolGen getConstantPool() {
        return this.cp;
    }
    
    public Field[] getFields() {
        return this.fieldList.toArray(Field.EMPTY_ARRAY);
    }
    
    public String getFileName() {
        return this.fileName;
    }
    
    public String[] getInterfaceNames() {
        return this.interfaceList.toArray(ArrayUtils.EMPTY_STRING_ARRAY);
    }
    
    public int[] getInterfaces() {
        final int size = this.interfaceList.size();
        final int[] interfaces = new int[size];
        Arrays.setAll(interfaces, i -> this.cp.addClass(this.interfaceList.get(i)));
        return interfaces;
    }
    
    public JavaClass getJavaClass() {
        final int[] interfaces = this.getInterfaces();
        final Field[] fields = this.getFields();
        final Method[] methods = this.getMethods();
        Attribute[] attributes = null;
        if (this.annotationList.isEmpty()) {
            attributes = this.getAttributes();
        }
        else {
            final Attribute[] annAttributes = AnnotationEntryGen.getAnnotationAttributes(this.cp, this.getAnnotationEntries());
            attributes = new Attribute[this.attributeList.size() + annAttributes.length];
            this.attributeList.toArray(attributes);
            System.arraycopy(annAttributes, 0, attributes, this.attributeList.size(), annAttributes.length);
        }
        final ConstantPool cp = this.cp.getFinalConstantPool();
        return new JavaClass(this.classNameIndex, this.superclassNameIndex, this.fileName, this.major, this.minor, super.getAccessFlags(), cp, interfaces, fields, methods, attributes);
    }
    
    public int getMajor() {
        return this.major;
    }
    
    public Method getMethodAt(final int pos) {
        return this.methodList.get(pos);
    }
    
    public Method[] getMethods() {
        return this.methodList.toArray(Method.EMPTY_ARRAY);
    }
    
    public int getMinor() {
        return this.minor;
    }
    
    public String getSuperclassName() {
        return this.superClassName;
    }
    
    public int getSuperclassNameIndex() {
        return this.superclassNameIndex;
    }
    
    @Override
    public int hashCode() {
        return ClassGen.bcelComparator.hashCode(this);
    }
    
    public void removeAttribute(final Attribute a) {
        this.attributeList.remove(a);
    }
    
    public void removeField(final Field f) {
        this.fieldList.remove(f);
    }
    
    public void removeInterface(final String name) {
        this.interfaceList.remove(name);
    }
    
    public void removeMethod(final Method m) {
        this.methodList.remove(m);
    }
    
    public void removeObserver(final ClassObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void replaceField(final Field old, final Field newField) {
        if (newField == null) {
            throw new ClassGenException("Replacement method must not be null");
        }
        final int i = this.fieldList.indexOf(old);
        if (i < 0) {
            this.fieldList.add(newField);
        }
        else {
            this.fieldList.set(i, newField);
        }
    }
    
    public void replaceMethod(final Method old, final Method newMethod) {
        if (newMethod == null) {
            throw new ClassGenException("Replacement method must not be null");
        }
        final int i = this.methodList.indexOf(old);
        if (i < 0) {
            this.methodList.add(newMethod);
        }
        else {
            this.methodList.set(i, newMethod);
        }
    }
    
    public void setClassName(final String name) {
        this.className = Utility.pathToPackage(name);
        this.classNameIndex = this.cp.addClass(name);
    }
    
    public void setClassNameIndex(final int classNameIndex) {
        this.classNameIndex = classNameIndex;
        this.className = Utility.pathToPackage(this.cp.getConstantPool().getConstantString(classNameIndex, (byte)7));
    }
    
    public void setConstantPool(final ConstantPoolGen constantPool) {
        this.cp = constantPool;
    }
    
    public void setMajor(final int major) {
        this.major = major;
    }
    
    public void setMethodAt(final Method method, final int pos) {
        this.methodList.set(pos, method);
    }
    
    public void setMethods(final Method[] methods) {
        this.methodList.clear();
        Collections.addAll(this.methodList, methods);
    }
    
    public void setMinor(final int minor) {
        this.minor = minor;
    }
    
    public void setSuperclassName(final String name) {
        this.superClassName = Utility.pathToPackage(name);
        this.superclassNameIndex = this.cp.addClass(name);
    }
    
    public void setSuperclassNameIndex(final int superclassNameIndex) {
        this.superclassNameIndex = superclassNameIndex;
        this.superClassName = Utility.pathToPackage(this.cp.getConstantPool().getConstantString(superclassNameIndex, (byte)7));
    }
    
    private AnnotationEntryGen[] unpackAnnotations(final Attribute[] attrs) {
        final List<AnnotationEntryGen> annotationGenObjs = new ArrayList<AnnotationEntryGen>();
        for (final Attribute attr : attrs) {
            if (attr instanceof RuntimeVisibleAnnotations) {
                final RuntimeVisibleAnnotations rva = (RuntimeVisibleAnnotations)attr;
                rva.forEach(a -> annotationGenObjs.add(new AnnotationEntryGen(a, this.getConstantPool(), false)));
            }
            else if (attr instanceof RuntimeInvisibleAnnotations) {
                final RuntimeInvisibleAnnotations ria = (RuntimeInvisibleAnnotations)attr;
                ria.forEach(a -> annotationGenObjs.add(new AnnotationEntryGen(a, this.getConstantPool(), false)));
            }
        }
        return annotationGenObjs.toArray(AnnotationEntryGen.EMPTY_ARRAY);
    }
    
    public void update() {
        if (this.observers != null) {
            for (final ClassObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
    
    static {
        ClassGen.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(final Object o1, final Object o2) {
                final ClassGen THIS = (ClassGen)o1;
                final ClassGen THAT = (ClassGen)o2;
                return Objects.equals(THIS.getClassName(), THAT.getClassName());
            }
            
            @Override
            public int hashCode(final Object o) {
                final ClassGen THIS = (ClassGen)o;
                return THIS.getClassName().hashCode();
            }
        };
    }
}
